/*
    SPDX-FileCopyrightText: 2021 Vlad Zahorodnii <vlad.zahorodnii@kde.org>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "highlightwindow.h"

namespace KWin
{

KWIN_EFFECT_FACTORY(HighlightWindowEffect,
                    "metadata.json.stripped")

} // namespace KWin

#include "main.moc"
