/*
    SPDX-FileCopyrightText: 2020 Vlad Zahorodnii <vlad.zahorodnii@kde.org>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "waylandshellintegration.h"

namespace KWin
{

WaylandShellIntegration::WaylandShellIntegration(QObject *parent)
    : QObject(parent)
{
}

} // namespace KWin

#include "moc_waylandshellintegration.cpp"
