/*
    SPDX-FileCopyrightText: 2022 David Redondo <kde@david-redondo.de>

    SPDX-License-Identifier: GPL-2.0-only OR GPL-3.0-only OR LicenseRef-KDE-Accepted-GPL
*/

#include "outputlocator.h"

namespace KWin
{
KWIN_EFFECT_FACTORY(OutputLocatorEffect, "metadata.json.stripped");
}

#include "main.moc"
