/*
    SPDX-FileCopyrightText: 2021 Vlad Zahorodnii <vlad.zahorodnii@kde.org>
    SPDX-FileCopyrightText: 2023 Andrew Shark <ashark at linuxcomp.ru>

    SPDX-License-Identifier: GPL-2.0-or-later
*/

#include "mousemark.h"

namespace KWin
{

KWIN_EFFECT_FACTORY(MouseMarkEffect,
                    "metadata.json.stripped")

} // namespace KWin

#include "main.moc"
